#include "iffimage.h"

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

typedef char byte;
typedef int int32;
typedef short int16;
typedef unsigned char uByte;
typedef unsigned int uInt32;
typedef unsigned short uInt16;
typedef float Float32;
typedef union _uIntFloat32 {
	uInt32 unt;
	Float32 flt;
} uIntFloat32;
// #define __IFF_DEBUG_

#define RGB_FLAG		 (1)
#define ALPHA_FLAG	 (2)
#define ZBUFFER_FLAG (4)

#define CHUNK_STACK_SIZE (32)

/*
 * Error code definitions 
 */
#define IFF_NO_ERROR		 (0)
#define IFF_OPEN_FAILS	 (1)
#define IFF_READ_FAILS	 (2)
#define IFF_BAD_TAG			(3)
#define IFF_BAD_COMPRESS (4)
#define IFF_BAD_STACK		(5)
#define IFF_BAD_CHUNK		(6)

/*
 * Define the IFF tags we are looking for in the file. 
 */
const uInt32 IFF_TAG_CIMG = ('C' << 24) | ('I' << 16) | ('M' << 8) | ('G');
const uInt32 IFF_TAG_FOR4 = ('F' << 24) | ('O' << 16) | ('R' << 8) | ('4');
const uInt32 IFF_TAG_TBHD = ('T' << 24) | ('B' << 16) | ('H' << 8) | ('D');
const uInt32 IFF_TAG_TBMP = ('T' << 24) | ('B' << 16) | ('M' << 8) | ('P');
const uInt32 IFF_TAG_RGBA = ('R' << 24) | ('G' << 16) | ('B' << 8) | ('A');
const uInt32 IFF_TAG_CLPZ = ('C' << 24) | ('L' << 16) | ('P' << 8) | ('Z');
const uInt32 IFF_TAG_ESXY = ('E' << 24) | ('S' << 16) | ('X' << 8) | ('Y');
const uInt32 IFF_TAG_ZBUF = ('Z' << 24) | ('B' << 16) | ('U' << 8) | ('F');
const uInt32 IFF_TAG_BLUR = ('B' << 24) | ('L' << 16) | ('U' << 8) | ('R');
const uInt32 IFF_TAG_BLRT = ('B' << 24) | ('L' << 16) | ('R' << 8) | ('T');
const uInt32 IFF_TAG_HIST = ('H' << 24) | ('I' << 16) | ('S' << 8) | ('T');

/*
 * For the stack of chunks 
 */
typedef struct _iff_chunk {
	uInt32 tag;
	uInt32 start;
	uInt32 size;
	uInt32 chunkType;
} iff_chunk;

iff_chunk chunkStack[CHUNK_STACK_SIZE];
int chunkDepth = -1;

/*
 * The current error state. 
 */
unsigned iff_error = IFF_NO_ERROR;

// -- Function prototypes, local to this file.
iff_chunk iff_begin_read_chunk(FILE * file);
void iff_end_read_chunk(FILE * file);
uByte *iff_read_data(FILE * file, int size);
uByte *iff_decompress_rle(FILE * file, uInt32 numBytes,
			  uByte * compressedData,
			  uInt32 compressedDataSize, uInt32 * compressedIndex);
uByte *iff_decompress_tile_rle(FILE * file, uInt16 width,
			       uInt16 height, uInt16 depth,
			       uByte * compressedData,
			       uInt32 compressedDataSize);
uByte *iff_read_uncompressed_tile(FILE * file, uInt16 width,
				  uInt16 height, uInt16 depth);

/*
 * -- Basic input functions.
 *
 */

uInt16 iff_get_short(FILE * file)
{
	uByte buf[2];
	size_t result = 0;

	result = fread(buf, 2, 1, file);

	if (result != 1) {
		if (iff_error == IFF_NO_ERROR) {
			iff_error = IFF_READ_FAILS;
		}
		return (0);
	}

	return ((buf[0] << 8) + (buf[1]));
}

uInt32 iff_get_long(FILE * file)
{
	uByte buffer[4];

	size_t result = fread(buffer, 4, 1, file);
	if (result != 1) {
		if (iff_error == IFF_NO_ERROR) {
			iff_error = IFF_READ_FAILS;
		}
		return (0);
	}

	return ((buffer[0] << 24) + (buffer[1] << 16)
		+ (buffer[2] << 8) + (buffer[3] << 0));
}

Float32 iff_get_float(FILE * file)
{
	uByte buffer[4];

	uIntFloat32 value;
	size_t result = fread(buffer, 4, 1, file);
	if (result != 1) {
		if (iff_error == IFF_NO_ERROR) {
			iff_error = IFF_READ_FAILS;
		}
		return (0.0);
	}

	value.unt = (buffer[3] << 24) + (buffer[2] << 16)
	    + (buffer[1] << 8) + (buffer[0] << 0);

	return (value.flt);
}

/*
 * IFF Chunking Routines.
 *
 */

iff_chunk iff_begin_read_chunk(FILE * file)
{
	chunkDepth++;
	if ((chunkDepth >= CHUNK_STACK_SIZE) || (chunkDepth < 0)) {
		if (iff_error == IFF_NO_ERROR) {
			iff_error = IFF_BAD_STACK;
		}
		return (chunkStack[0]);
	}

	chunkStack[chunkDepth].start = ftell(file);
	chunkStack[chunkDepth].tag = iff_get_long(file);
	chunkStack[chunkDepth].size = iff_get_long(file);

	if (chunkStack[chunkDepth].tag == IFF_TAG_FOR4) {
		// -- We have a form, so read the form type tag as well. 
		chunkStack[chunkDepth].chunkType = iff_get_long(file);
	} else {
		chunkStack[chunkDepth].chunkType = 0;
	}

#ifdef __IFF_DEBUG_
	printf("Beginning Chunk: %c%c%c%c",
	       (((&chunkStack[chunkDepth].tag)[0] >> 24) & 0xFF),
	       (((&chunkStack[chunkDepth].tag)[0] >> 16) & 0xFF),
	       (((&chunkStack[chunkDepth].tag)[0] >> 8) & 0xFF),
	       (((&chunkStack[chunkDepth].tag)[0] >> 0) & 0xFF));

	printf("	start: %lu", chunkStack[chunkDepth].start);
	printf("	size: %lu", chunkStack[chunkDepth].size);
	if (chunkStack[chunkDepth].chunkType != 0) {
		printf("	type: %c%c%c%c",
		       (((&chunkStack[chunkDepth].chunkType)[0] >> 24) &
			0xFF),
		       (((&chunkStack[chunkDepth].chunkType)[0] >> 16) &
			0xFF),
		       (((&chunkStack[chunkDepth].chunkType)[0] >> 8) &
			0xFF),
		       (((&chunkStack[chunkDepth].chunkType)[0] >> 0) & 0xFF));
	}
	printf("	depth: %d\n", chunkDepth);
#endif

	return (chunkStack[chunkDepth]);
}

void iff_end_read_chunk(FILE * file)
{
	uInt32 end =
	    chunkStack[chunkDepth].start + chunkStack[chunkDepth].size + 8;
	int part;

	if (chunkStack[chunkDepth].chunkType != 0) {
		end += 4;
	}
	// Add padding 
	part = end % 4;
	if (part != 0) {
		end += 4 - part;
	}

	fseek(file, end, SEEK_SET);

#ifdef __IFF_DEBUG_
	printf("Closing Chunk: %c%c%c%c\n\n",
	       (((&chunkStack[chunkDepth].tag)[0] >> 24) & 0xFF),
	       (((&chunkStack[chunkDepth].tag)[0] >> 16) & 0xFF),
	       (((&chunkStack[chunkDepth].tag)[0] >> 8) & 0xFF),
	       (((&chunkStack[chunkDepth].tag)[0] >> 0) & 0xFF));
#endif

	chunkDepth--;
}

uByte *iff_read_data(FILE * file, int size)
{
	uByte *buffer = (uByte *) malloc(size * sizeof(uByte));
	size_t result = 0;

#ifdef __IFF_DEBUG_
	printf("read_data	size: %d\n", size);
#endif

	result = fread(buffer, size, 1, file);

	if (result <= 0) {
		if (iff_error == IFF_NO_ERROR) {
			iff_error = IFF_READ_FAILS;
		}
		return (0);
	}

	return (buffer);
}

/*
 * Compression Routines 
 *
 */

uByte *iff_decompress_rle(FILE * file, uInt32 numBytes,
			  uByte * compressedData,
			  uInt32 compressedDataSize, uInt32 * compressedIndex)
{

	uByte *data = (uByte *) malloc(numBytes * sizeof(uByte));
	uByte nextChar, count;
	int i;
	uInt32 byteCount = 0;

	for (i = 0; i < (int)numBytes; ++i) {
		data[i] = 0;
	}

#ifdef __IFF_DEBUG_
	printf("Decompressing data %d\n", numBytes);
#endif

	while (byteCount < numBytes) {

		if (*compressedIndex >= compressedDataSize) {
			break;
		}

		nextChar = compressedData[*compressedIndex];
		(*compressedIndex)++;

		count = (nextChar & 0x7f) + 1;
		if ((byteCount + count) > numBytes)
			break;

		if (nextChar & 0x80) {

			// We have a duplication run

			nextChar = compressedData[*compressedIndex];
			(*compressedIndex)++;

			// assert( ( byteCount + count ) <= numBytes ); 
			for (i = 0; i < count; ++i) {
				data[byteCount] = nextChar;
				byteCount++;
			}
		} else {
			// We have a verbatim run 
			for (i = 0; i < count; ++i) {

				data[byteCount] =
				    compressedData[*compressedIndex];
				(*compressedIndex)++;
				byteCount++;
			}
		}
		assert(byteCount <= numBytes);
	}

	return (data);
}

uByte *iff_decompress_tile_rle(FILE * file,
			       uInt16 width, uInt16 height, uInt16 depth,
			       uByte * compressedData,
			       uInt32 compressedDataSize)
{
	if (compressedData == 0)
		return 0;
	uByte *channels[4];
	uByte *data;
	int i, k, row, column;
	uInt32 compressedIndex = 0;

#ifdef __IFF_DEBUG_
	printf("Decompressing tile [ %hd, ", width);
	printf("%hd, ", height);
	printf("%hd ]\n", depth);
#endif

	// -- Decompress the different channels (RGBA)
	// -- ERROR CHECK, MUST OPERATE ON RGBA !!!
	if (depth != 4) {
		if (iff_error == IFF_NO_ERROR) {
			iff_error = IFF_BAD_COMPRESS;
		}
		return (0);
	}

	for (i = (depth - 1); i >= 0; --i) {
		channels[i] =
		    iff_decompress_rle(file, width * height, compressedData,
				       compressedDataSize, &compressedIndex);
	}

	// -- Pack all of the channels from the decompression into an RGBA
	// array.
	data = (uByte *) malloc(width * height * depth * sizeof(uByte));

	for (row = 0; row < height; row++) {
		for (column = 0; column < width; column++) {
			for (k = 0; k < depth; k++) {
				data[depth * (row * width + column) + k] =
				    channels[k][row * width + column];
			}
		}
	}

	for (i = 0; i < depth; i++) {
		free(channels[i]);
	}

	return (data);
}

uByte *iff_read_uncompressed_tile(FILE * file,
				  uInt16 width, uInt16 height, uInt16 depth)
{
	uByte *data = 0;
	uByte pixel[4];
	int i, j, d, index;
	size_t res;
	data = (uByte *) malloc(width * height * depth * sizeof(uByte));

#ifdef __IFF_DEBUG_
	printf("Begin reading uncompressed tile\n", NULL);
#endif

	for (i = 0; i < height; i++) {
		index = i * width * depth;
		for (j = 0; j < width; ++j) {
			res = fread(pixel, depth, 1, file);
			for (d = (depth - 1); d >= 0; --d) {
				data[index] = pixel[d];
				++index;
			}
		}
	}

#ifdef __IFF_DEBUG_
	printf("End reading uncompressed tile\n", NULL);
#endif

	return (data);
}

iff_image *iff_load(const char *filename)
{
	iff_chunk chunkInfo;
	iff_image *image;

	// -- Header info.
	uInt32 width, height, depth, npixels;
	uInt32 flags, compress;
	uInt16 tiles;
	uInt16 tbmpfound;

	uInt16 x1, x2, y1, y2, tile_width, tile_height;
	uInt32 tile;
	uInt32 ztile;

	uInt32 iscompressed, i;
	uByte *tileData;
	uInt32 remainingDataSize;

	uInt32 oldSpot;
	uInt32 fileLength;

	FILE *file;

	// -- Initialize the top of the chunk stack.
	chunkDepth = -1;
	image = 0;

	// -- Open the file.  
	file = fopen(filename, "rb");
	if (!file) {
		if (iff_error == IFF_NO_ERROR) {
			iff_error = IFF_OPEN_FAILS;
		}
		return (0);
	}
	// -- File should begin with a FOR4 chunk of type CIMG
	chunkInfo = iff_begin_read_chunk(file);
	if (chunkInfo.chunkType != IFF_TAG_CIMG) {
		if (iff_error == IFF_NO_ERROR) {
			// -- This is not a CIMG, it is not an IFF Image.
			iff_error = IFF_BAD_TAG;
		}
		fclose(file);
		return (0);
	}

	/*
	 * Read the image header
	 * OK, we have a FOR4 of type CIMG, look for the following tags
	 *              FVER    
	 *              TBHD    bitmap header, definition of size, etc.
	 *              AUTH
	 *              DATE
	 */
	while (1) {

		chunkInfo = iff_begin_read_chunk(file);

		// -- Right now, the only info we need about the image is in TBHD
		// -- so search this level until we find it.
		if (chunkInfo.tag == IFF_TAG_TBHD) {

			// -- Header chunk found
			width = iff_get_long(file);
			height = iff_get_long(file);
			iff_get_short(file);	// -- Don't support 
			iff_get_short(file);	// -- Don't support 
			flags = iff_get_long(file);
			iff_get_short(file);	// -- Don't support
			tiles = iff_get_short(file);
			compress = iff_get_long(file);

#ifdef __IFF_DEBUG_
			printf("****************************************\n");
			printf("Width: %u\n", width);
			printf("Height: %u\n", height);
			printf("flags: 0x%X\n", flags);
			printf("tiles: %hu\n", tiles);
			printf("compress: %u\n", compress);
			printf("****************************************\n");
#endif

			iff_end_read_chunk(file);

			if (compress > 1) {
				if (iff_error == IFF_NO_ERROR) {
					iff_error = IFF_BAD_COMPRESS;
				}
				fclose(file);
				return (0);
			}

			break;
		} else {

#ifdef __IFF_DEBUG_
			// Skipping unused data at FOR4 <size> CIMG depth
			printf("Skipping Chunk: %c%c%c%c\n",
			       (((&chunkStack[chunkDepth].
				  tag)[0] >> 24) & 0xFF),
			       (((&chunkStack[chunkDepth].
				  tag)[0] >> 16) & 0xFF),
			       (((&chunkStack[chunkDepth].tag)[0] >> 8) & 0xFF),
			       (((&chunkStack[chunkDepth].
				  tag)[0] >> 0) & 0xFF));
#endif

			iff_end_read_chunk(file);
		}
	}			/* END find TBHD while loop */

	// -- Number of channels.
	depth = 0;

	if (flags & RGB_FLAG) {
		depth += 3;
	}

	if (flags & ALPHA_FLAG) {
		depth += 1;
	}

	npixels = width * height;

	// -- Allocate the image struct.
	image = (iff_image *) malloc(sizeof(iff_image));
	image->width = width;
	image->height = height;
	image->depth = depth;
	image->znear = 0.0;
	image->zfar = 0.0;
	image->zesx = 0.0;
	image->zesy = 0.0;
	image->rgba = 0;
	image->zbuffer = 0;
	image->blurvec = 0;

	image->rgba = (uByte *) malloc(npixels * depth * sizeof(uByte));
	// memset( image->rgba, 0, npixels * depth * sizeof( uByte ) );

	if (flags & ZBUFFER_FLAG) {
		image->zbuffer = (Float32 *) malloc(npixels * sizeof(Float32));
		// memset( image->zbuffer, 0, npixels * sizeof( Float32 ) );
	}
	// -- Assume the next FOR4 of type TBMP
	tbmpfound = 0;

	// Read the tiled image data
	while (!tbmpfound) {

		uInt16 tile_area;
		chunkInfo = iff_begin_read_chunk(file);

		/*
		 * OK, we have a FOR4 of type TBMP, (embedded FOR4)
		 * look for the following tags
		 *              RGBA    color data,     RLE compressed tiles of 32 bbp data
		 *              ZBUF    z-buffer data, 32 bit float values
		 *              CLPZ    depth map specific, clipping planes, 2 float values
		 *              ESXY    depth map specific, eye x-y ratios, 2 float values
		 *              HIST    
		 *              VERS
		 *              FOR4 <size>     BLUR (twice embedded FOR4)
		 */
		if (chunkInfo.chunkType == IFF_TAG_TBMP) {
			tbmpfound = 1;

			// Image data found
			tile = 0;
			ztile = 0;

#ifdef __IFF_DEBUG_
			printf("Reading image tiles\n");
#endif

			if (!(flags & ZBUFFER_FLAG)) {
				ztile = tiles;
			}

			if (depth == 0) {
				tile = tiles;
			}
			// -- Read tiles
			while ((tile < tiles) || (ztile < tiles)) {

				chunkInfo = iff_begin_read_chunk(file);

				if (!(chunkInfo.tag == IFF_TAG_RGBA)
				    && !(chunkInfo.tag == IFF_TAG_ZBUF)) {
					if (iff_error == IFF_NO_ERROR) {
						iff_error = IFF_BAD_CHUNK;
					}
					iff_free(image);
					fclose(file);
					return (0);
				}
				// Get tile size and location info
				x1 = iff_get_short(file);
				y1 = iff_get_short(file);
				x2 = iff_get_short(file);
				y2 = iff_get_short(file);

				remainingDataSize = chunkInfo.size - 8;

				tile_width = x2 - x1 + 1;
				tile_height = y2 - y1 + 1;
				tile_area = tile_width * tile_height;

#ifdef __IFF_DEBUG_
				printf("Tile x1: %hu	", x1);
				printf("y1: %hu	", y1);
				printf("x2: %hu	", x2);
				printf("y2: %hu\n", y2);
#endif

				if ((unsigned int)chunkInfo.size >=
				    (tile_width * tile_height * depth + 8)) {
					// -- Compression was not used for this tile.
					iscompressed = 0;
				} else {
					iscompressed = 1;
				}

				// -- OK, we found an RGBA chunk, eat it.
				if (chunkInfo.tag == IFF_TAG_RGBA) {
					if (depth == 0) {
						iff_end_read_chunk(file);
						continue;
					}

					if (iscompressed) {
						uByte *data =
						    iff_read_data(file,
								  remainingDataSize);
						tileData =
						    iff_decompress_tile_rle
						    (file, tile_width,
						     tile_height, depth, data,
						     remainingDataSize);
						free(data);
					} else {
						tileData =
						    iff_read_uncompressed_tile
						    (file, tile_width,
						     tile_height, depth);
					}

					if (tileData) {

						/*
						 * Dump RGBA data to our data structure 
						 */
						uInt16 i;

						// -- base is the tile offset into the packed
						// array
						uInt32 base =
						    image->depth *
						    (image->width * y1 + x1);

						for (i = 0; i < tile_height;
						     i++) {
							memcpy(&image->
							       rgba[base +
								    image->
								    depth * i *
								    width],
							       &tileData[depth *
									 i *
									 tile_width],
							       tile_width *
							       depth *
							       sizeof(uByte));
						}

						/*
						 * End RGBA dump 
						 */

						free(tileData);
						tileData = 0;
					}

					iff_end_read_chunk(file);
					tile++;
				}
				/*
				 * END RGBA chunk 
				 */
				// -- OK, we found a ZBUF chunk, eat it....hmmm, tasty
				else if (chunkInfo.tag == IFF_TAG_ZBUF) {

					// -- Read in the tile data.
					if (iscompressed) {
						uByte *data =
						    iff_read_data(file,
								  remainingDataSize);
						tileData =
						    iff_decompress_tile_rle
						    (file, tile_width,
						     tile_height, 4, data,
						     remainingDataSize);
						free(data);
					} else {
						tileData =
						    iff_read_uncompressed_tile
						    (file, tile_width,
						     tile_height, 4);
					}

					if (tileData) {
						/*
						 * Dump DEPTH data into our structure of floats
						 */

						int i, j, base;

						for (i = 0; i < tile_height;
						     i++) {
							for (j = 0;
							     j < tile_width;
							     j++) {

								uInt32 index;
								uIntFloat32
								    value;
								index =
								    4 * (i *
									 tile_width
									 + j);

								base =
								    y1 * width +
								    x1;

								/*
								 * If MSB of float is 1, it is negative
								 * as it should be, so let's process it
								 * if it's not 1, just set the depth to 0
								 *
								 * NOTE: depth values are stored as
								 * -1/z in light eye space in the IFF file.
								 */

								// if( tileData[index+3] & 0x80 ) {

								value.unt =
								    (tileData
								     [index +
								      3] << 24)
								    +
								    (tileData
								     [index +
								      2] << 16)
								    +
								    (tileData
								     [index +
								      1] << 8) +
								    (tileData
								     [index] <<
								     0);

								image->
								    zbuffer[base
									    +
									    i *
									    width
									    +
									    j] =
								    value.flt;
								// ( -1.0f * (*( (Float32 *)&value ) ));

								// } else {
								// image->zbuffer[base + i*width+j] = 0.0;
								// }

							}	/* End DEPTH dump */
						}

						free(tileData);
					}

					iff_end_read_chunk(file);
					ztile++;

				}
				/*
				 * END ZBUF chunk 
				 */
			}	/* END while TBMP tiles */

		}
		/*
		 * END if TBMP 
		 */
		else {

#ifdef __IFF_DEBUG_
			// Skipping unused data IN THE BEGINNING OF THE FILE
			printf("Skipping Chunk in search of TBMP: %c%c%c%c\n",
			       (((&chunkStack[chunkDepth].
				  tag)[0] >> 24) & 0xFF),
			       (((&chunkStack[chunkDepth].
				  tag)[0] >> 16) & 0xFF),
			       (((&chunkStack[chunkDepth].tag)[0] >> 8) & 0xFF),
			       (((&chunkStack[chunkDepth].
				  tag)[0] >> 0) & 0xFF));
#endif

			iff_end_read_chunk(file);
		}
	}

	oldSpot = ftell(file);
	fileLength = 0;
	fseek(file, 0, SEEK_END);
	fileLength = ftell(file);
	fseek(file, oldSpot, SEEK_SET);

	while (1) {

		chunkInfo = iff_begin_read_chunk(file);

		if (chunkInfo.tag == IFF_TAG_CLPZ) {

			image->znear = iff_get_float(file);
			image->zfar = iff_get_float(file);

#ifdef __IFF_DEBUG_
			printf("Got clipping info: %f %f\n", image->znear,
			       image->zfar);
#endif

			iff_end_read_chunk(file);

		}
		/*
		 * END CLPZ chunk 
		 */
		else if (chunkInfo.tag == IFF_TAG_ESXY) {
			image->zesx = iff_get_float(file);
			image->zesy = iff_get_float(file);

#ifdef __IFF_DEBUG_
			printf("Got esxy info: %f %f\n", image->zesx,
			       image->zesy);
#endif

			iff_end_read_chunk(file);
		}
		/*
		 * END ESXY chunk 
		 */
		else if (chunkInfo.tag == IFF_TAG_FOR4) {

			if (chunkInfo.chunkType == IFF_TAG_BLUR) {

				// -- FIXME: GET THE BLUR INFO HERE
				if (image->blurvec) {
					free(image->blurvec);
				}

				while (1) {

					chunkInfo = iff_begin_read_chunk(file);

					if (chunkInfo.tag == IFF_TAG_BLRT) {

						// read in values, uncompressed and in a linear
						// sort
						// of manner, uh huh...
						/*
						 * printf( "%d\n", iff_get_long( file ) ); printf( 
						 * "%d\n", iff_get_long( file ) ); printf( "%d\n", 
						 * iff_get_long( file ) ); printf( "%d\n",
						 * iff_get_long( file ) ); 
						 */
						iff_get_long(file);	// -- Don't know what
						// these are
						iff_get_long(file);
						iff_get_long(file);
						iff_get_long(file);

						image->blurvec =
						    (Float32 *) malloc(npixels *
								       2 *
								       sizeof
								       (Float32));

						for (i = 0; i < npixels; i++) {
							image->blurvec[2 * i] =
							    iff_get_float(file);
							image->blurvec[2 * i +
								       1] =
							    iff_get_float(file);
						}

						iff_end_read_chunk(file);
						break;
					}

					else {
#ifdef __IFF_DEBUG_
						printf
						    ("Skipping Chunk in search of BLRT: %c%c%c%c\n",
						     (((&chunkStack[chunkDepth].
							tag)[0] >> 24) & 0xFF),
						     (((&chunkStack[chunkDepth].
							tag)[0] >> 16) & 0xFF),
						     (((&chunkStack[chunkDepth].
							tag)[0] >> 8) & 0xFF),
						     (((&chunkStack[chunkDepth].
							tag)[0] >> 0) & 0xFF));
#endif

						iff_end_read_chunk(file);
					}
				}
#ifdef __IFF_DEBUG_
				printf("Found FOR4 BLUR\n");
#endif
			}

			iff_end_read_chunk(file);
		}

		else {

#ifdef __IFF_DEBUG_
			// Skipping unused data IN THE BEGINNING OF THE FILE
			printf("Skipping Chunk in search of CLPZ: %c%c%c%c\n",
			       (((&chunkStack[chunkDepth].
				  tag)[0] >> 24) & 0xFF),
			       (((&chunkStack[chunkDepth].
				  tag)[0] >> 16) & 0xFF),
			       (((&chunkStack[chunkDepth].tag)[0] >> 8) & 0xFF),
			       (((&chunkStack[chunkDepth].
				  tag)[0] >> 0) & 0xFF));
#endif

			iff_end_read_chunk(file);
		}

		if ((width * height + ftell(file)) > fileLength) {

#ifdef __IFF_DEBUG_
			printf("End of parsable data, time to quit\n");
#endif

			break;
		}

	}

	fclose(file);

	return (image);
}

void iff_free(iff_image * image)
{
	if (image) {

		if (image->rgba) {
			free(image->rgba);
		}
		if (image->zbuffer) {
			free(image->zbuffer);
		}
		if (image->blurvec) {
			free(image->blurvec);
		}

		free(image);

		image = 0;
	}
}

unsigned iff_get_error()
{
	unsigned err = iff_error;
	iff_error = IFF_NO_ERROR;
	return (err);
}

const char *iff_error_string(unsigned iff_errno)
{
	switch (iff_errno) {
	case (IFF_NO_ERROR):
		return ("no error");
	case (IFF_OPEN_FAILS):
		return ("cannot open file");
	case (IFF_READ_FAILS):
		return ("cannot read file");
	case (IFF_BAD_TAG):
		return ("unexpected tag");
	case (IFF_BAD_COMPRESS):
		return ("unknown compression format");
	case (IFF_BAD_STACK):
		return ("tag stack corrupt");
	case (IFF_BAD_CHUNK):
		return ("unexpected chunk");
	}

	return ("");
}
